from AutoRPE.UtilsRPE.BasicStructures import Interface
from AutoRPE.UtilsRPE.BasicFunctions import combine_types, remove_literals, is_all_inside_parenthesis, close_brackets
from AutoRPE.UtilsRPE.LoadAndStore import load_vault, list_of_intrinsics
from AutoRPE.UtilsRPE.Cleaner import clean_contents
from AutoRPE.UtilsRPE.AttributeParser import split_attributes
from AutoRPE.UtilsRPE.Splitter import split_hardcoded_array, split_elements_in_comparison, split_elements_in_operations
from AutoRPE.UtilsRPE.CallManager import is_call_to_function, find_call_arguments, find_called_function
from AutoRPE.UtilsRPE.NatureDeterminer import has_operations, is_a_number, has_intrinsics

import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern

import re
import warnings


def retrieve_variable(contents, routine=None, vault=None):

    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions
    if not vault:
        vault = load_vault()

    contents = contents.replace(" ", "").lower().strip()

    try:
        purged_contents = clean_from_intrinsics(contents)
        contents = purged_contents
    except ValueError:
        print("Not able to clean content ", contents, " from intrinsics")

    contents = clean_contents(contents, vault)

    if re.search(r"^\w+$", contents):
        try:
            return vault.get_variable(contents, routine=routine)
        except Error.VariableNotFound:
           raise Error.VariableNotFound
    if contents.count("%val"):
        contents = contents.replace("%val", "")
    if contents.count("%"):
        try:
            return get_structure_member(contents, routine, vault=vault)
        except Error.MemberNotFound:
            raise Error.VariableNotFound

    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        return retrieve_variable(contents, routine=routine, vault=vault)

    raise Error.VariableNotFound("Did not find this variable: %s" % contents)


def get_structure_member(contents, routine, vault):
    # Retrieve variable corresponding to a structure member
    contents = contents.lower()
    # Let's check if the structure has been declared in this same routine
    variable_name = contents.split("%")[0]
    variable = [v for v in routine.variables + routine.module.variables if v.name.lower() == variable_name]
    if not variable:
        try:
            variable = [[var for var in m.accessible_var if var.name.lower() == variable_name][0]
                        for m in vault.modules if m.name == routine.module.name.lower()][0]
        except IndexError:
            raise Error.VariableNotFound

        routine = variable.routine
    if variable:
        return get_member_from_root(contents, routine, vault)
    else:
        return get_member_from_tail(contents, routine, vault)


###########CHECK THIS IS ACTUALLY USED
def get_member_from_tail(contents, routine, vault):
    structure_member = contents.split("%")[-1]
    coincidences = [v for v in vault.get_variable(structure_member) if v.is_member_of is not None]
    if len(coincidences) == 1:
        # Just one structure has such a member
        return coincidences[0]
    else:
        # More than one structure has such a member
        for match in coincidences:
            index = 2
            member_of = match.is_member_of
            # Checks the different member going backward, and looks at the compatibility with available structures
            while index <= len(contents.split("%")):
                structure_member = contents.split("%")[-index]
                # Structures which has member of the same type
                struct = [v for v in vault.get_variable(structure_member) if v.type == member_of]
                # No structures correspond to this match, go to the next
                if len(struct) == 0:
                    break
                if len(struct) == 1:
                    # Just one structure matches, but we need to check that the parent structure type coincides
                    if struct[0].is_member_of is not None:
                        member_of = struct[0].is_member_of
                        index = index + 1
                    # One match is found and there is no parent structure
                    else:
                        return match
                # More than one struct is found, check parent
                if len(struct) > 1:
                    # Reached the root member we still have more than a match: check if the routine coincide
                    if index == len(contents.split("%")):
                        struct_filter = [s for s in struct if s.routine == routine]
                        if len(struct_filter) == 1:
                            return match
                    index = index + 1
        # Something went wrong in the search
        raise Error.MemberNotFound


def get_member_from_root(contents, routine, vault):
    # Retrieve variable corresponding to a structure member
    root = contents.split("%")[0]
    root_variable = [v for v in routine.variables + routine.module.variables if v.name.lower() == root][0]
    try:
        root_derived_type = [dt for dt in vault.derived_types if dt.name.lower() == root_variable.type.lower()][0]
    except IndexError:
        raise Error.DerivedTypeNotFound
    layers = contents.split("%")[1:]
    for l in layers:
        try:
            layer = [v for v in root_derived_type.variables if v.name.lower() == l.lower()][0]
        except IndexError:
            # It means that mother type is an extended type
            try:
                layer = [v for v in root_derived_type.extension.variables if v.name.lower() == l.lower()][0]
            except IndexError:
                raise Error.MemberNotFound
            if layer:
                pass
            else:
                raise IndexError("Variable %s not found", contents)

        try:
            root_derived_type = [dt for dt in vault.derived_types if dt.name == layer.type][0]
        except IndexError:
            if l == contents.split("%")[-1]:
                pass
            else:
                raise IndexError("Variable %s not found", contents)

    return layer


def get_type_of_contents(contents, routine=None, vault=None):
    # TODO: Simplify this function
    # I should do a first filter at this point distinguishing if its a string, a simple variable ("\w+")
    # if its a function/array (is_call_to_function is True), is a keyword argument, or other exceptions
    if contents == "*":
        return None
    if not vault:
        vault = load_vault()
    if contents[0] == "-" or contents[0] == "+":
        contents = contents[1:]

    try:
        contents = split_hardcoded_array(contents)[0]
        return get_type_of_contents(contents, routine=routine, vault=vault)
    except Error.IsNotAnArray:
        pass

    if is_all_inside_parenthesis(contents):
        contents = contents.strip()[1:-1]
        # raise AssertionError("Remove the first and last bracket")

    # Case with call to a function with array arguments after call foo(a(1:3), b)
    match = re.search(r"(\w* *\([^=]*\)) *(\(.*\))", contents)
    if match:
        contents = match.group(1)

    contents = contents.lower().strip()
    if contents[0] in ["'", '"'] or contents[:4] == 'trim':
        return "char"

    is_a_comparison = split_elements_in_comparison(contents)
    if is_a_comparison is not None:
        return "logical"

    contents = contents.replace(" ", "").strip()

    if is_call_to_function(contents):
        fun_name = find_called_function(contents)
        if fun_name in vault.intrinsics:
            if fun_name in ["reshape", "min", "max", "sum", "log", "log10", "maxval", "minval", "abs", "sqrt", "maxloc",
                            "minloc", "huge", "tiny", "transfer", "sign"]:
                try:
                    return get_type_of_contents(find_call_arguments(contents)[0], routine=routine, vault=vault)
                except RecursionError as rec:
                    print('Recursion error: {}'.format(rec.args[0]))
            elif fun_name in ["int", "nint", "count", "size", "mod", "umask", "len", "len_trim", "count"]:
                return "integer"
            elif fun_name in ["real", "float", "sin", "cos", "acos", "exp", "cosh"]:
                if fun_name == "real":
                    return get_type_of_real_cast(string=contents)

                return VariablePrecision.wp
            elif fun_name in ["cmplx"]:
                return "complex"
            elif fun_name in ["trim", "ctime", "adjustl"]:
                return "char"
            elif fun_name in ["present"]:
                return "logical"
            elif fun_name in ["null", "c_loc"]:
                return "pointer"
            elif fun_name in ["rpe"]:
                return "rpe_var"
            else:
                raise ValueError("This intrinsic should be considered too: %s" % fun_name)
        else:
            try:
                fun = vault.get_procedure_by_name(fun_name)
                if isinstance(fun, Interface):
                    return fun.procedures[0].type
                else:
                    return fun.type
            except Error.ProcedureNotFound:
                pass
            # warnings.warn("If its a function should be on the library or an intrinsic, isn't it? %s|nLooking if its a variable... " % contents)
            # return type_of_contents(fun_name, routine=routine, vault=vault)

    if has_operations(contents):
        split_elements = split_elements_in_operations(contents, routine=routine, vault=vault)
        try:
            types = [get_type_of_contents(element, routine, vault) for element in split_elements]
            return combine_types(types)
        except RecursionError as rec:
            print('Recursion error: {}'.format(rec.args[0]))
    if is_keyword_argument(contents):
        # Removing before first =
        contents = contents[contents.find("=") + 1:].strip()
        try:
            return get_type_of_contents(contents, routine=routine, vault=vault)
        except RecursionError as rec:
            print('Recursion error: {}'.format(rec.args[0]))

    contents = clean_contents(contents, vault)

    if re.match("^" + RegexPattern.number_exp_DPE_notation + "$", contents):
        return VariablePrecision.wp

    if contents.find("nf90_") == 0:
        return "integer"
    if contents.find("set_nf90_") == 0:
        return "integer"

    if contents.count("'") or contents.count("\""):
        return "char"
    if contents.count(","):
        # If there are commas at this point, it must correspond with a temporal array
        # Array elements should have the same type, so we will assume that all the elements
        # have the same type than the first element
        first_element = split_attributes(contents)[0]
        return get_type_of_contents(first_element, vault=vault)
    if contents.count(","):
        raise ValueError("Argument should not contain commas at this point %s" % contents)

    if contents.count("(") or contents.count(")"):
        warnings.warn("WARNING: Argument should not contain brackets at this point % s" % contents)
        contents = contents.replace("(", "")
        contents = contents.replace(")", "")

    if contents.count(".false.") or contents.count(".true."):
        return "logical"

    # Checks if it is a number
    num_type = is_a_number(contents)
    if num_type:
        return num_type


    if vault:
        abs_content = contents
        # in case a variable's name is preceded from unary operator
        if contents[0] == '-' or contents[0] == '+':
            abs_content = contents[1:]

        if contents.count("%"):
            # TODO: Strange way of handling errors. Check if there's a better way of doing it
            if contents.split("%")[-1] == "val":
                return VariablePrecision.dp

            member = get_structure_member(contents, routine, vault=vault)
            return member.type
            # return type_of_derived_type_element(contents, routine=routine, vault=vault)

        if re.search(r"^\w+$", abs_content):
            try:
                var = vault.get_variable(abs_content, routine=routine)
            except Error.VariableNotFound:
                raise Error.VariableNotFound
            return var.type


    if contents.count(">") or contents.count("<") or contents.count("=="):
        return "logical"

    if vault:
        if contents.count("mpi_"):
            return "integer"
        elif contents == "rpe":
            return "rpe_var"
        else:
            warnings.warn("Not able to find type! :(  |%s|" % contents)
    else:
        pass

    warnings.warn("Did not find type of: %s" % contents)
    # raise AssertionError("Did not find type of: %s" % contents)


# TODO: Improve this function to work properly in all cases.
def dimension_of_contents(contents, routine=None, vault=None):
    contents = remove_literals(contents)
    contents = contents.lower()
    # it is a Char
    if contents.count("'") or contents.count('"') or contents.lower().count("trim"):
        return 0
    if contents.count(".false.") or contents.count(".true."):
        return 0

    # Matches arrays in the form of (/a,b,c/) or [a,b,c]
    array_match = re.search(RegexPattern.hardcoded_array, contents)
    if array_match:
        return 1
    # Adding a workaround for casts
    if is_call_to_function(contents):
        fun = find_called_function(contents)
        if fun == "real":
            contents = find_call_arguments(contents)[0]
        elif fun == "sum" or fun == "maxval":
            if contents.count("dim="):
                dim = dimension_of_contents(find_call_arguments(contents)[0], routine=routine, vault=vault)
                if dim:
                    return dim - 1
                else:
                    return dim
            else:
                contents = find_call_arguments(contents)[0]
        elif fun in ["sqrt", "min", "max"]:
            contents = find_call_arguments(contents)[0]
        else:
            # it was not a cast?
            pass
    if has_operations(contents):
        elements = split_elements_in_operations(contents)
        dimensions = [dimension_of_contents(e, routine=routine, vault=vault) for e in elements]
        dimensions = [d for d in dimensions if d is not None]
        if dimensions == []:
            dimensions = [0]
        return max(dimensions)
    if is_a_number(contents):
        return 0
    if contents.count(":"):
        return contents.count(":")
        # Case with call to a function with array arguments after

    # Arrays with index specified
    if re.search(r"\w *\([0-9\w]+\)", contents):
        return 0
    try:
        var = retrieve_variable(contents, routine=routine, vault=vault)
        # Checking if we have the var + slice specification... if it is a slice and does not have any colon
        # then it must be an scalar
        if re.search(r"%s *\(.*\)" % var.name, contents):
            return 0
        return var.dimension
    except AssertionError:
        return None


def type_of_derived_type_element(contents, routine=None, derived_type=None, vault=None):
    if not vault:
        vault = load_vault()

    struct = contents[:contents.find("%")]
    var = contents[contents.find("%") + 1:]
    if var == "val":
        return VariablePrecision.wp

    if not derived_type:
        derived_type = get_type_of_contents(struct, routine=routine, vault=vault)

    if derived_type is None:
        raise AssertionError("Did not type of %s" % contents)

    if var.count("%"):

        struct = var[:var.find("%")]
        derived_type = type_of_derived_type_element(struct, routine=routine, derived_type=derived_type, vault=vault)
        var = var[var.find("%") + 1:]
        if var == "val":
            return VariablePrecision.wp
        dt = vault.get_derived_type(derived_type)
        try:
            v = dt.get_variable(var)
        except AttributeError:
            pass

        try:
            return v.type
        except AttributeError:
            pass
    else:
        if vault:
            dt = vault.get_derived_type(derived_type)
            try:
                return dt.get_variable(var).type
            except AttributeError:
                pass

        return "unknown"


def match_arguments_and_interface(arguments, interface, routine=None, vault=None):
    argument_types = [get_type_of_contents(argument, routine=routine, vault=vault) for argument in arguments]
    argument_dimensions = [dimension_of_contents(argument, routine=routine, vault=vault) for argument in arguments]
    keyword_arguments = [_x.split("=")[0].strip() if is_keyword_argument(_x) else False for _x in arguments]
    return match_argument_info_and_interface(argument_types, interface=interface,
                                             argument_dimensions=argument_dimensions,
                                             keyword_arguments=keyword_arguments, vault=vault)


def match_argument_info_and_interface(argument_types, interface, argument_dimensions=None, keyword_arguments=None,
                                      vault=None):
    # This function is intended to take the argument info (types and dimensions) and return the subroutine
    # inside the interface that matches it.
    number_of_arguments = len(argument_types)
    if keyword_arguments is None:
        keyword_arguments = [False for _x in argument_types]
    for procedure in interface.procedures:
        if not vault:
            dummy_arguments = procedure.get_dummy_arguments()
        else:
            dummy_arguments = [var for var in procedure.variables if var.is_dummy_argument and not var.is_member_of]

        try:
            mandatory_arguments = [argument for argument in dummy_arguments if not argument.is_optional]
        except AttributeError:
            print("A problem with dummy argument in procedure %s", procedure.name)
        if len(dummy_arguments) < number_of_arguments:
            continue
        if len(mandatory_arguments) > number_of_arguments:
            continue
        match = True
        for index, argument_type in enumerate(argument_types):
            if match is False:
                continue
            keyword = keyword_arguments[index]
            if keyword:
                var = procedure.get_variable_by_name(keyword)
                if var:
                    dummy_type = var.type
                    dummy_dimension = var.dimension
                else:
                    dummy_type = None
                    dummy_dimension = None
                    match = False
            else:
                dummy_type = procedure.get_variable_by_position(index).type
                dummy_dimension = procedure.get_variable_by_position(index).dimension
            if dummy_type != argument_type:
                if (dummy_type in VariablePrecision.real_types and argument_type == "rpe_var") or (
                        dummy_type == "rpe_var" and argument_type in VariablePrecision.real_types):
                    pass
                else:
                    match = False
            if argument_dimensions is not None:
                # At this point we'll just ignore char variables.
                if argument_type == "char":
                    continue
                argument_dimension = argument_dimensions[index]
                if argument_dimension is not None and argument_dimension != dummy_dimension:
                    match = False

        if match:
            return procedure
    raise AssertionError("No matching routine found for interface : %s" % interface.name)


def is_keyword_argument(argument):
    pattern = "^[^()'\"=]*=[^=]"
    if re.search(pattern, argument):
        return True
    else:
        return False


def get_type_of_real_cast(string):
    # Leave only the arguments
    arguments = find_call_arguments(string)
    if len(arguments) == 1:
        return VariablePrecision.wp
    elif len(arguments) == 2:
        argument, type = arguments
        return VariablePrecision.lookup_table[type]

##PRIVATE
# Try and remove intrinsics from around a variable
def clean_from_intrinsics(contents, vault=None):
    if vault:
        intrinsics = vault.intrinsics
    else:
        intrinsics = list_of_intrinsics
    pattern = r"(\w+) *\("
    m = [_m for _m in re.finditer(pattern, contents)]
    ret_val = contents
    if m:
        for mat in reversed(m):
            function = contents[mat.start(1):mat.end(1)]
            if function in intrinsics:
                if function == "null":
                    continue
                arg_of_func = contents[mat.end(0):]

                try:
                    arg_of_intrinsic = close_brackets(arg_of_func)
                    # Bad workaround to fix the case in which the argument is not a vector
                    if arg_of_intrinsic.count(",("):
                        arg_of_intrinsic = arg_of_intrinsic.split(",")[0]
                    intrinsic_call = close_brackets(contents[mat.start(1):])
                except ValueError:
                    return contents
                ret_val = contents.replace(intrinsic_call, arg_of_intrinsic)
                if has_intrinsics(ret_val):
                    try:
                        ret_val = clean_from_intrinsics(ret_val)
                    except RecursionError:
                        return contents
                else:
                    return ret_val
            elif function == "real":
                arguments = find_call_arguments(contents)
                return arguments[0]
    return ret_val